# CheapBulletWhizz release 1 by David Bowland
# ./addons/eventscripts/cheapbulletwhizz/cheapbulletwhizz.py

# >>> To configure this addon please see cheapbulletwhizz.cfg <<<

"""
Provides bullet whizz sounds. Requires only ES 2.0+
"""


import es
import math
import os.path
import random

import psyco
psyco.full()


info = es.AddonInfo()
info.name     = 'CheapBulletWhizz'
info.version  = '1'
info.url      = 'http://addons.eventscripts.com/addons/view/cheapbulletwhizz'
info.basename = 'cheapbulletwhizz'
info.author   = 'SuperDave'


dict_events  = {'bullet_whizz':{'userid':'short', 'distance':'float', 'attacker':'short'}}
dict_options = {'cheapbulletwhizz_distance_min':[50, 'Distance from a bullet\'s trajectory at which players hear whizz sound at maximum volume'], 'cheapbulletwhizz_distance_max':[300, 'Maximum distance a bullet\'s trajectory can be to hear a whizz sound'], 'cheapbulletwhizz_opponentsonly':[0, '0 = players hear the whizz from all bullets, 1 = players hear the whizz from bullets of opponents only']}


players    = {}
sounds     = []
last_sound = {}
noisy_set  = False


class BulletWhizzManager:
   userid       = 0
   last_fire    = ()
   hurt_players = set()

   def __init__(self, userid):
      self.userid = userid

      self.reset()

   def reset(self):
      """Resets bullet stats"""
      self.last_fire    = ()

      self.hurt_players = set((self.userid,))

   def weaponFire(self):
      """Records the origin of the bullet"""
      location = list(es.getplayerlocation(self.userid))
      location[2] += 45 if es.getplayerprop(self.userid, 'CBasePlayer.localdata.m_Local.m_bDucked') % 2 else 60

      self.last_fire    = tuple(location)

      self.hurt_players = set((self.userid,))

   def hurt(self, userid):
      """Marks the player as hurt by the bullet"""
      self.hurt_players.add(userid)

   def bulletImpact(self, impact_location):
      """
      Calculates the slop and distance of the shot
      Loops through a filtered list of players and retrieves each player's distance to the bullet's trajectory
      Player the whizz sound and fires bullet_whizz
      """
      if not self.last_fire: return

      impact_slopes   = self._getSlopes(self.last_fire, impact_location)
      impact_distance = sum((self.last_fire[x] - impact_location[x]) ** 2 for x in range(3)) ** 0.5

      distance_min = float(es.ServerVar('cheapbulletwhizz_distance_min'))
      distance_max = float(es.ServerVar('cheapbulletwhizz_distance_max'))

      if int(es.ServerVar('cheapbulletwhizz_opponentsonly')):
         team    = es.getplayerteam(self.userid)
         userids = filter(lambda x: es.getplayerteam(x) == 5 - team, es.getUseridList())

      else:
         userids = es.getUseridList()

      for userid in filter(lambda x: not es.getplayerprop(x, 'CBasePlayer.pl.deadflag'), set(userids).difference(self.hurt_players)):
         player_location = list(es.getplayerlocation(userid))
         player_location[2] += 45 if es.getplayerprop(userid, 'CBasePlayer.localdata.m_Local.m_bDucked') % 2 else 60

         player_distance = sum((player_location[x] - self.last_fire[x]) ** 2 for x in range(3)) ** 0.5
         player_impact   = sum((player_location[x] - impact_location[x]) ** 2 for x in range(3)) ** 0.5

         angles_to_player = self._getAngles(impact_slopes, self._getSlopes(self.last_fire, player_location))
         distance_to_shot = sum((math.sin(angles_to_player[x]) * player_distance) ** 2 for x in range(3)) ** 0.5

         max_dist = (impact_distance ** 2 + distance_to_shot ** 2) ** 0.5
         if max_dist <= player_distance or max_dist <= player_impact:
            continue

         volume = 1.0 - (distance_to_shot - distance_min) / distance_max
         if volume > 0:
            if sounds:
               es.playsound(userid, self._getSound(userid), volume if volume < 1 else 1)

            es.event('initialize', 'bullet_whizz')
            es.event('setint', 'bullet_whizz', 'userid', userid)
            es.event('setfloat', 'bullet_whizz', 'distance', distance_to_shot)
            es.event('setint', 'bullet_whizz', 'attacker', self.userid)
            es.event('fire', 'bullet_whizz')

   def _getAngles(self, slopes1, slopes2):
      """Returns the angles between two slopes in a 3-element tuple"""
      return tuple(math.atan(abs(slopes1[x] - slopes2[x]) / (1 + slopes1[x] * slopes2[x])) for x in range(3))

   def _getSlopes(self, point1, point2):
      """Returns the slope between two points in a 3-element tuple"""
      return tuple(self.__getSingleSlope(point1[x], point1[y], point2[x], point2[y]) for x, y in ((0, 1), (1, 2), (0, 2)))

   def __getSingleSlope(self, x1, y1, x2, y2):
      """Returns the slope between two points"""
      return ((y2 - y1) / (x2 - x1)) if x2 - x1 else 0

   def _getSound(self, userid):
      """Returns a whizz sound avoiding the last sound if applicable"""
      global last_sound

      choices = filter(lambda x: x <> last_sound[userid], sounds) if last_sound.has_key(userid) else ()
      if not choices:
         choices = sounds

      sound = random.choice(choices)
      last_sound[userid] = sound

      return sound


def load():
   """
   Ensures critical server variables are created by the config
   Calls es_map_start
   Creates a BulletWhizzManager instance for each player
   Sets eventscripts_noisy to 1
   """
   global players
   global noisy_set

   for str_option in dict_options:
      es.ServerVar(str_option, dict_options[str_option][0], dict_options[str_option][1])
   if os.path.isfile(es.getAddonPath('cheapbulletwhizz') + '/cheapbulletwhizz.cfg'):
      es.server.cmd('es_xmexec ../addons/eventscripts/cheapbulletwhizz/cheapbulletwhizz.cfg')
   else:
      es.dbgmsg(0, 'CheapBulletWhizz: Unable to load cheapbulletwhizz.cfg! Please ensure it is in the ./cheapbulletwhizz/ directory.')

   es_map_start({})

   for userid in es.getUseridList():
      players[userid] = BulletWhizzManager(userid)

   noisy = es.ServerVar('eventscripts_noisy')
   es.doblock('corelib/noisy_on')
   if not int(noisy):
      noisy_set = True
      noisy.set(1)


def es_map_start(event_var):
   """
   Calls load_events
   Marks whizz sounds for download
   Clears ditionary of last sounds played
   """
   load_events()

   path = str(es.ServerVar('eventscripts_gamedir')) + '/sound/'
   for sound in filter(lambda x: os.path.isfile(path + x), sounds):
      es.stringtable('downloadables', 'sound/' + sound)

   last_sound.clear()


def player_spawn(event_var):
   """
   Creates a BulletWhizzManager instance for the player
   Resets the BWM instance's bullet stats
   """
   global players

   userid = int(event_var['userid'])
   if not players.has_key(userid):
      players[userid] = BulletWhizzManager(userid)

   players[userid].reset()


def player_hurt(event_var):
   """Marks the player as hurt by the attacker"""
   attacker = int(event_var['attacker'])
   if players.has_key(attacker):
      players[attacker].hurt(int(event_var['userid']))


def weapon_fire(event_var):
   """Stores the player's position of weapon fire"""
   players[int(event_var['userid'])].weaponFire()


def bullet_impact(event_var):
   """Initiates whizz sounds"""
   players[int(event_var['userid'])].bulletImpact(tuple(float(event_var[x]) for x in ('x', 'y', 'z')))


def unload():
   """Reverts eventscripts_noisy"""
   if noisy_set:
      es.ServerVar('eventscripts_noisy').set(0)
   else:
      es.doblock('corelib/noisy_off')

   for str_option in dict_options:
      es.ServerVar(str_option).set(0)


def addsound_cmd():
   """
   cheapbulletwhizz_addsound <"sound relative to ./sound/">
   Adds a bullet whizz sound
   """
   if es.getargc() == 2:
      sounds.append(es.getargv(1))

   else:
      es.dbgmsg(0, 'Syntax: cheapbulletwhizz_addsound <\"sound relative to ./sound/\">')
if not es.exists('command', 'cheapbulletwhizz_addsound'):  
   es.regcmd('cheapbulletwhizz_addsound', 'cheapbulletwhizz/addsound_cmd', 'cheapbulletwhizz_addsound <\"sound relative to ./sound/\">\nAdds a bullet whizz sound')


def removesound_cmd():
   """
   cheapbulletwhizz_removesound <"sound relative to ./sound/">
   Removes a bullet whizz sound
   """
   if es.getargc() == 2:
      sound = es.getargv(1)
      if sound in sounds: 
         sounds.remove(sound)

   else:
      es.dbgmsg(0, 'Syntax: cheapbulletwhizz_removesound <\"sound relative to ./sound/\">')
if not es.exists('command', 'cheapbulletwhizz_removesound'):  
   es.regcmd('cheapbulletwhizz_removesound', 'cheapbulletwhizz/removesound_cmd', 'cheapbulletwhizz_removesound <\"sound relative to ./sound/\">\nRemoves a bullet whizz sound')


def load_events():
   """Loads the .res file for the events"""
   if os.path.isfile(es.getAddonPath('cheapbulletwhizz') + '/cheapbulletwhizz.res'):
      es.loadevents('declare', 'addons/eventscripts/cheapbulletwhizz/cheapbulletwhizz.res')

   else:
      if not os.path.isfile(es.getAddonPath('cheapbulletwhizz') + 'es_cheapbulletwhizz_events_db.txt'):
         es.keygroupcreate('cheapbulletwhizz_events')
         for str_event in dict_events:
            es.keycreate('cheapbulletwhizz_events', str_event)
            for str_event_var in dict_events[str_event]:
               es.keysetvalue('cheapbulletwhizz_events', str_event, str_event_var, dict_events[str_event][str_event_var])
         es.keygroupsave('cheapbulletwhizz_events', '|cheapbulletwhizz')
         es.keygroupdelete('cheapbulletwhizz_events')
      es.loadevents('declare', 'addons/eventscripts/cheapbulletwhizz/es_cheapbulletwhizz_events_db.txt')
      es.dbgmsg(0, 'CheapBulletWhizz: Unable to load cheapbulletwhizz.res! Please ensure it is in the ./cheapbulletwhizz/ directory.')